package main

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"os"
	"path"
	"strings"
	"sync"
	"time"

	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/urfave/cli"
	"gitlab.lehouerou.net/laurent/dnsupdater/internal/config"
	"gitlab.lehouerou.net/laurent/dnsupdater/internal/dns"
	"gitlab.lehouerou.net/laurent/dnsupdater/internal/dns/gandi"
	"gitlab.lehouerou.net/laurent/dnsupdater/internal/outboundip"
	"gitlab.lehouerou.net/laurent/dnsupdater/internal/outboundip/ddwrt"
	"gopkg.in/gomail.v2"
)

var (
	configuration config.Configuration
	configPath    string
)

func main() {

	output := zerolog.ConsoleWriter{Out: os.Stdout, TimeFormat: time.RFC3339}
	output.FormatLevel = func(i interface{}) string {
		return strings.ToUpper(fmt.Sprintf("| %-6s|", i))
	}
	output.FormatFieldName = func(i interface{}) string {
		return fmt.Sprintf("%s:", i)
	}
	output.FormatFieldValue = func(i interface{}) string {
		return strings.ToUpper(fmt.Sprintf("%s", i))
	}
	log.Logger = zerolog.New(output).With().Timestamp().Logger()

	app := cli.NewApp()
	app.Name = "dnsupdater"
	app.Usage = "Automatically update dns"
	app.Before = before
	app.Action = start
	app.Flags = []cli.Flag{

		cli.StringFlag{
			Name:   "config, C",
			Usage:  "path to config file",
			Value:  "/config",
			EnvVar: "CONFIG_PATH",
		},
	}

	if err := app.Run(os.Args); err != nil {
		log.Fatal().Err(err).Send()
	}
}

func before(c *cli.Context) error {
	configPath = path.Join(c.String("config"), "conf.json")
	file, err := os.Open(configPath)
	if err != nil {
		return fmt.Errorf("error while opening config file: %v", err)
	}
	defer file.Close()
	decoder := json.NewDecoder(file)
	configuration = config.Configuration{}
	err = decoder.Decode(&configuration)
	if err != nil {
		return err
	}
	return nil
}

func start(*cli.Context) error {
	ctx := context.Background()

	httpClient := &http.Client{
		Timeout: 30 * time.Second,
	}
	outboundIpRequester := ddwrt.NewIpRequester(
		httpClient,
		configuration.RouterStatusUrl,
		configuration.RouterLogin,
		configuration.RouterPassword)

	dnsmanager := gandi.NewDnsManager(
		httpClient,
		configuration.GandiKey)

	if err := update(ctx, outboundIpRequester, dnsmanager); err != nil {
		SendStatusMail(fmt.Sprintf("Error during update process : %s", err))
		return err
	}
	return nil
}

func update(ctx context.Context, requester outboundip.IpRequester, manager dns.Manager) error {
	newIpResult := <-requester.GetOutboundIp()

	if newIpResult.Error != nil {
		return fmt.Errorf("getting outbound ip: %v", newIpResult.Error)
	}
	newIp := newIpResult.Ip

	wg := sync.WaitGroup{}

	updatedDomains := ""
	updatedDomainsChannel := make(chan string)
	for _, domain := range configuration.Domains {
		domain := domain
		wg.Add(1)
		go func() {
			defer wg.Done()
			updated, err := checkDomain(manager, domain, newIp)
			if err != nil {
				log.Error().Err(err).Str("domain", domain).Msg("checking domain ip change")
			}
			if updated {
				updatedDomainsChannel <- domain
			}
		}()
	}
	go func() {
		wg.Wait()
		close(updatedDomainsChannel)
	}()
	for {
		select {
		case <-ctx.Done():
			return errors.New("context canceled")
		case d, ok := <-updatedDomainsChannel:
			if !ok {
				updatedDomainsChannel = nil
				break
			}
			updatedDomains += fmt.Sprintf("\t- %s\n", d)
		}
		if updatedDomainsChannel == nil {
			break
		}
	}
	if updatedDomains != "" {
		SendStatusMail(fmt.Sprintf(`Home IP has changed : %s
Dns updated successfully for domains
%s`, newIp, updatedDomains))
	}
	return nil
}

func checkDomain(manager dns.Manager, domainname string, outboundip string) (bool, error) {
	currentIp, err := manager.GetARecord(domainname)
	if err != nil {
		return false, fmt.Errorf("getting current ip: %v", err)
	}

	if currentIp == outboundip {
		return false, nil
	}
	log.
		Info().
		Str("domain", domainname).
		Str("current ip", currentIp).
		Str("new ip", outboundip).
		Msg("ip has changed")
	err = manager.SetARecord(domainname, outboundip)
	if err != nil {
		return false, fmt.Errorf("setting current ip: %v", err)
	}
	return true, nil
}

func SendStatusMail(messageText string) {

	settings := configuration.MailSettings
	if settings == nil {
		log.Warn().Msg("no smtp settings defined > status mail not sent")
		return
	}

	d := gomail.NewDialer(settings.SmtpHost, settings.SmtpPort, settings.SmtpLogin, settings.SmtpPassword)
	s, err := d.Dial()
	if err != nil {
		log.Err(err).Msg("dialing smtp server")
		return
	}

	m := gomail.NewMessage()
	for _, r := range settings.To {
		m.SetHeader("From", settings.Sender)
		m.SetAddressHeader("To", r, "")
		m.SetHeader("Subject", "DnsUpdater Status")
		m.SetBody("text/plain", messageText)

		if err := gomail.Send(s, m); err != nil {
			log.Warn().
				Err(err).
				Str("recipient", r).
				Msg("could not send email")
		}
		m.Reset()
	}
}
